#include <AsyncWebSocket.h>

// Import required libraries
#include <WiFi.h>
#include <AsyncTCP.h>
#include <ESPAsyncWebServer.h>
#include <max6675.h>
#include <Adafruit_SH110X.h>
#include <Adafruit_GFX.h>
#include <Wire.h>
#include <OneWire.h>
#include <DallasTemperature.h>
#include "SPIFFS.h"
#include "FS.h"

char bufferTemp[64];
AsyncWebSocketClient* firstClient = NULL; // Store the first connected client

// Set to true to define Relay as Normally Open (NO)
#define RELAY_NO    true

// Set number of relays
#define NUM_RELAYS  2

// Assign each GPIO to a relay
int relayGPIOs[NUM_RELAYS] = {4, 23};

// Replace with your network credentials
const char* ssid     = "Home_main";
const char* password = "DHA128MNi@zi";

bool relayState = 0;
const int ledPin = 2;

// Create AsyncWebServer object on port 80
AsyncWebServer server(80);
AsyncWebSocket ws("/ws");

#define i2c_Address 0x3c
#define SCREEN_WIDTH 128
#define SCREEN_HEIGHT 64
#define OLED_RESET -1
Adafruit_SH1106G display = Adafruit_SH1106G(SCREEN_WIDTH, SCREEN_HEIGHT, &Wire, OLED_RESET);

//DS18x20
#define ONE_WIRE_BUS 2 //Data wire is plug to D2

// Setup a oneWire instance to communicate with any OneWire devices
OneWire oneWire(ONE_WIRE_BUS);

// Setup Temperature senson
DallasTemperature sensors(&oneWire);

//K-type
int thermoDO = 19; // SO of  MAX6675 module to D19
int thermoCS = 5; // CS of MAX6675 module to D5
int thermoCLK = 18; // SCK of MAX6675 module to D18

MAX6675 thermocouple(thermoCLK, thermoCS, thermoDO);

void handleWebSocketMessage(void *arg, uint8_t *data, size_t len) {
    AwsFrameInfo *info = (AwsFrameInfo*)arg;
    if (info->final && info->index == 0 && info->len == len && info->opcode == WS_TEXT) {
        data[len] = 0;
        String message = String((char*)data);
        Serial.print("message: ");
        Serial.println(message);

        if (message.startsWith("toggle")) {
            relayState = !relayState;
            ws.textAll(String(relayState));
            Serial.print("relayState: ");
            Serial.println(relayState);
        } else if (message.startsWith("range")) {
            int indexOfRange = message.indexOf("range");
            if (indexOfRange != -1) {
                int rangeValue = message.substring(indexOfRange + 5).toInt();
                ws.textAll("range" + String(rangeValue));
                // Save the range value to temp.txt
                writeFile(SPIFFS, "/temp.txt", String(rangeValue));
            }
        }
    }
}


void eventHandler(AsyncWebSocket *server, AsyncWebSocketClient *client, AwsEventType type, void *arg, uint8_t *data, size_t len) {
  switch (type) {
    case WS_EVT_CONNECT:
      Serial.printf("WebSocket client #%u connected from %s\n", client->id(), client->remoteIP().toString().c_str());

      // If this is the first connected client, assign it to firstClient
      if (!firstClient) {
        firstClient = client;
      }
      
      break;
    case WS_EVT_DISCONNECT:
      Serial.printf("WebSocket client #%u disconnected\n", client->id());

      // If the first connected client disconnects, reset firstClient
      if (firstClient == client) {
        firstClient = NULL;
      }
      
      break;
    case WS_EVT_DATA:
      handleWebSocketMessage(arg, data, len);
      digitalWrite(relayGPIOs[1], relayState);
      break;
    case WS_EVT_PONG:
      Serial.println("PONG received from first connected");
      break;
    case WS_EVT_ERROR:
      break;
  }
}

String processor(const String& var){ 
  if(var == "CHECK"){
    return relayState ? "checked" : "";
  }
  if (var == "FIRE_TEMP_PLACEHOLDER") {
    String fireTemp = "";
    fireTemp = thermocouple.readCelsius();
    return fireTemp;
  }
  if(var == "WATER_TEMP_PLACEHOLDER"){
    String waterTemp = "";
    waterTemp = getWaterTemperature();
    return waterTemp;
  }
  if (var == "TEMP_VALUE") {
    readFile(SPIFFS, "/temp.txt");
    return bufferTemp;
  }
  return String();
}

void writeFile(fs::FS &fs, const char * path, String message){
    Serial.printf("Writing file: %s\r\n", path);

    File file = fs.open(path, FILE_WRITE);
    if(!file){
        Serial.println("- failed to open file for writing");
        return;
    }
    if(file.print(message)){
        Serial.println("- file written");
    } else {
        Serial.println("- write failed");
    }
    file.close();
}

void readFile(fs::FS &fs, const char * path){
    Serial.printf("Reading file: %s\r\n", path);

    File file = fs.open(path);
    if(!file || file.isDirectory()){
        Serial.println("- failed to open file for reading");
        return;
    }

    Serial.println("- read from file:");
    
    while (file.available()) {
       int l = file.readBytesUntil('\n', bufferTemp, sizeof(bufferTemp));
       bufferTemp[l] = 0;
       Serial.println(bufferTemp);
    }
    file.close();
}

void setup(){
  // Serial port for debugging purposes
  Serial.begin(115200);
  
  if(!SPIFFS.begin(true)){
    Serial.println("SPIFFS Mount Failed");
    return;
  }
  display.begin(i2c_Address, true);
  display.clearDisplay();
  delay(100);

  // Set all relays to off when the program starts - if set to Normally Open (NO), the relay is off when you set the relay to HIGH
  for(int i=1; i<=NUM_RELAYS; i++){
    pinMode(relayGPIOs[i-1], OUTPUT);
    digitalWrite(relayGPIOs[i-1], HIGH);
  }
  
  pinMode(ledPin, OUTPUT);
  digitalWrite(ledPin, LOW);

  Serial.printf("Connecting to %s ", ssid);
  // Connect to Wi-Fi
  WiFi.begin(ssid, password);
  while (WiFi.status() != WL_CONNECTED) {
    delay(1000);
    Serial.print(".");
  }

  Serial.printf("\nWiFi connected..!\nIP Address: ");
  // Print ESP32 Local IP Address
  Serial.println(WiFi.localIP());

  ws.onEvent(eventHandler);
  server.addHandler(&ws);

  // Route for root / web page
  server.on("/", HTTP_GET, [](AsyncWebServerRequest *request){
              request->send(SPIFFS, "/index.html", "text/html", false, processor);
            });
  server.on("/style.css", HTTP_GET, [](AsyncWebServerRequest *request){
              request->send(SPIFFS, "/style.css", "text/css"); 
            });

  server.on("/socket.js", HTTP_GET, [](AsyncWebServerRequest *request){
              request->send(SPIFFS, "/socket.js", "text/javascript"); 
            });

  // Start server
  server.begin();
  Serial.println("HTTP server started");

  display.setTextSize(1);
  display.setTextColor(SH110X_WHITE);
  display.setCursor(0,0);
  display.print("Connected To: ");
  display.display();

  display.setTextSize(1);
  display.setTextColor(SH110X_WHITE);
  display.setCursor(0,10);
  display.print(ssid);
  display.display();

  display.setTextSize(1);
  display.setTextColor(SH110X_WHITE);
  display.setCursor(0,20);
  display.print("IP: ");
  display.setCursor(20,20);
  display.print(WiFi.localIP());
  display.display();

  sensors.begin();
  readFile(SPIFFS, "/temp.txt");  
}

void loop() {
  // Check if there are active WebSocket clients
  if (ws.count() > 0) {
    static unsigned long lastTemperatureUpdate = 0;
    static unsigned long lastPingTime = 0;
    unsigned long currentMillis = millis();
  
    // Throttle temperature updates to once every 10 second
    if (currentMillis - lastTemperatureUpdate >= 10000) {
      // Read the temperature
      float fireTemp = thermocouple.readCelsius();
      float waterTemp = getWaterTemperature();
      
      // Send the temperature to all WebSocket clients
      ws.textAll("fireTemp" + String(fireTemp));
      ws.textAll("waterTemp" + String(waterTemp));    
      lastTemperatureUpdate = currentMillis;
    }
  }

  // Periodically send a ping to the first connected client
  if (firstClient){
    if(currentMillis - lastPingTime >= 30000) {
      firstClient->ping();
      lastPingTime = currentMillis;
    }
  }
  // rest code
  ws.cleanupClients();
}

float getWaterTemperature() {
  sensors.requestTemperatures();
  return sensors.getTempCByIndex(0);
}
